import dbt.exceptions
from contextlib import contextmanager
from dataclasses import dataclass
from dbt.adapters.base import Credentials
from dbt.adapters.sql import SQLConnectionManager
from dbt.logger import GLOBAL_LOGGER as logger
from dbt.contracts.connection import AdapterResponse, ConnectionState, AdapterRequiredConfig
from typing import Optional, Dict
from pyDLI.dli import Connection


@dataclass
class DLICredentials(Credentials):
    region: str
    project_id: str
    access_id: str
    secret_key: str
    queue: str
    database: str
    schema: str
    obs_endpoint: Optional[str] = None
    dli_endpoint: Optional[str] = None
    hints: Optional[Dict[str, str]] = None

    _ALIASES = {"ak": "access_id", "sk": "secret_key"}

    @property
    def type(self):
        """Return name of adapter."""
        return "dli"

    @property
    def unique_field(self):
        """
        Hashed and included in anonymous telemetry to track adapter adoption.
        Pick a field that can uniquely identify one team/organization building with this adapter
        """
        return self.region + "#" + self.database

    def _connection_keys(self):
        """
        List of keys to display in the `dbt debug` output.
        """
        return ("region", "project_id", "access_id", "database", "schema", "queue")


class DLIConnectionManager(SQLConnectionManager):
    TYPE = "dli"

    def __init__(self, profile: AdapterRequiredConfig):
        profile.query_comment.comment = None

        super().__init__(profile)

    @contextmanager
    def exception_handler(self, sql):
        try:
            yield
        except Exception as exc:
            logger.error("Error while running:\n{}".format(sql))
            logger.error(exc)

            if len(exc.args) == 0:
                raise

            thrift_resp = exc.args[0]
            if hasattr(thrift_resp, "status"):
                msg = thrift_resp.status.errorMessage
                raise dbt.exceptions.DbtRuntimeError(msg)
            else:
                raise dbt.exceptions.DbtRuntimeError(str(exc))

    @classmethod
    def open(cls, connection):
        """
        Receives a connection object and a Credentials object
        and moves it to the "open" state.
        """
        if connection.state == ConnectionState.OPEN:
            logger.debug("Connection is already open, skipping open.")
            return connection

        credentials: DLICredentials = connection.credentials
        try:
            if credentials.dli_endpoint is not None and credentials.dli_endpoint != '':
                host = "dli://%s/%s?queuename=%s&database=%s" % (
                    credentials.dli_endpoint, credentials.project_id,
                    credentials.queue, credentials.schema)
            else:
                host = "dli://dli.%s.myhuaweicloud.com/%s?queuename=%s&database=%s" % (
                    credentials.region, credentials.project_id,
                    credentials.queue, credentials.schema)
            kwargs = {
                "host": host,
                "ak": credentials.access_id,
                "sk": credentials.secret_key,
                "obs_endpoint": credentials.obs_endpoint,
            }
            handle = Connection(**kwargs)
            connection.state = "open"
            connection.handle = handle
        except Exception as exc:
            logger.error("Error opening connection: {}".format(exc))
            connection.handle = None
            connection.state = "fail"
        return connection

    @classmethod
    def get_response(cls, cursor) -> AdapterResponse:
        # dli does not support cursor and rowcount
        # https://github.com/dbt-labs/dbt-spark/issues/142
        message = "OK"
        return AdapterResponse(_message=message)

    def add_begin_query(self, *args, **kwargs):
        logger.debug("Not Supported: add_begin_query")

    def add_commit_query(self, *args, **kwargs):
        logger.debug("Not Supported: add_commit_query")

    def commit(self, *args, **kwargs):
        logger.debug("Not Supported: commit")

    def rollback(self, *args, **kwargs):
        logger.debug("Not Supported: rollback")

    def cancel(self, connection):
        connection.handle.cancel()
