import agate
from dataclasses import dataclass

import dbt
from dbt.adapters.sql import SQLAdapter
from dbt.adapters.base import AdapterConfig
from dbt.adapters.base.relation import BaseRelation, InformationSchema
from dbt.clients import agate_helper
from dbt.contracts.graph.manifest import Manifest
from dbt.logger import GLOBAL_LOGGER as logger
from typing import List, Set, cast, Optional, Dict, Any, Iterable

from dli.dli_client import DliClient
from typing_extensions import TypeAlias

from .colums import DLIColumn
from .connections import DLIConnectionManager, DLICredentials
from .relation import DLIRelation

LIST_SCHEMAS_MACRO_NAME = "list_schemas"
LIST_RELATIONS_MACRO_NAME = "list_relations_without_caching"
SHOW_CREATE_TABLE_MACRO_NAME = "show_create_table"
RENAME_RELATION_MACRO_NAME = "rename_relation"

KEY_TABLE_INFORMATION = "# Detailed Table Information"
KEY_TABLE_OWNER = "Owner:"

@dataclass
class DLIConfig(AdapterConfig):
    partitioned_by: Optional[List[Dict[str, str]]] = None
    properties: Optional[Dict[str, str]] = None


class DLIAdapter(SQLAdapter):
    """
    Controls actual implmentation of adapter, and ability to override certain methods.
    """

    ConnectionManager = DLIConnectionManager
    Relation: TypeAlias = DLIRelation
    Column: TypeAlias = DLIColumn
    AdapterSpecificConfigs: TypeAlias = DLIConfig

    @property
    def dli(self) -> DliClient:
        return self.connections.get_thread_connection().handle.dli_client

    @property
    def credentials(self) -> DLICredentials:
        return self.config.credentials

    @classmethod
    def date_function(cls) -> str:
        return "current_timestamp()"

    @classmethod
    def convert_number_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        # TODO CT-211
        decimals = agate_table.aggregate(agate.MaxPrecision(col_idx))  # type: ignore[attr-defined]
        return "double" if decimals else "bigint"

    @classmethod
    def convert_datetime_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "timestamp"

    @classmethod
    def convert_date_type(cls, agate_table, col_idx):
        return "date"

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        raise NotImplementedError("DLI does not support a native time type. Use a timestamp instead.")

    @classmethod
    def convert_text_type(cls, agate_table, col_idx: int) -> str:
        return "string"

    def timestamp_add_sql(self, add_to: str, number: int = 1, interval: str = "hour") -> str:
        """the interval could be one of [dd, mm, yyyy, mi, ss, year, month, mon, day, hour, hh]'"""
        return f"dateadd({add_to}, {number}, '{interval}')"

    def quote(self, identifier):
        return f"`{identifier}`"

    def check_schema_exists(self, database: str, schema: str) -> bool:
        return True

    def list_schemas(self, database: str) -> List[str]:
        results = self.execute_macro(LIST_SCHEMAS_MACRO_NAME, kwargs={"database": database})
        return [row[0] for row in results]

    def list_relations_without_caching(
            self,
            schema_relation: DLIRelation = None,
    ) -> List[DLIRelation]:
        # show views
        view_sql_job = self.dli.execute_sql(
            "show views from " + schema_relation.__str__(), queue_name=self.credentials.queue)
        rows = view_sql_job.get_result()
        views = []
        view_symbols = []
        while rows.next():
            schema = rows.fetch().columns[0]
            view = rows.fetch().columns[1]
            view_symbols.append(schema + "@" + view)
            views.append(DLIRelation.create(database=schema_relation.database, schema=schema, identifier=view, type="view"))

        # show tables
        sql_job = self.dli.execute_sql(
            "show tables from " + schema_relation.__str__(), queue_name=self.credentials.queue)
        rows = sql_job.get_result()
        tables = []
        while rows.next():
            schema = rows.fetch().columns[0]
            table = rows.fetch().columns[1]
            if schema + "@" + table in view_symbols:
                continue
            tables.append(DLIRelation.create(database=schema_relation.database, schema=schema, identifier=table, type="table"))
        return tables + views

    def get_columns_in_relation(self, relation: DLIRelation):
        try:
            rows: List[agate.Row] = super().get_columns_in_relation(relation)
            columns = self.parse_describe_formatted(relation, rows)
        except dbt.exceptions.DbtRuntimeError as e:
            # impala would throw error when table doesn't exist
            errmsg = getattr(e, "msg", "")
            if (
                    "Table or view not found" in errmsg
                    or "NoSuchTableException" in errmsg
                    or "Could not resolve path" in errmsg
                    or "Table not found" in errmsg
            ):
                return []
            else:
                raise e

        return columns

    def _get_columns_for_catalog(self, relation: DLIRelation) -> Iterable[Dict[str, Any]]:
        columns = self.get_columns_in_relation(relation)

        for column in columns:
            as_dict = column.to_column_dict()
            as_dict["column_name"] = as_dict.pop("column", None)
            as_dict["column_type"] = as_dict.pop("dtype")
            as_dict["table_database"] = None
            yield as_dict

    def parse_describe_formatted(
            self, relation: Relation, raw_rows: List[agate.Row]
    ) -> List[DLIColumn]:
        # Convert the Row to a dict
        dict_rows = [dict(zip(row._keys, row._values)) for row in raw_rows]
        # Find the separator between columns and partitions information
        # by the DESCRIBE EXTENDED {{relation}} statement
        partition_separator_pos = self.find_partition_information_separator(dict_rows)

        # Find the separator between the rows and the metadata provided
        # by the DESCRIBE EXTENDED {{relation}} statement
        table_separator_pos = self.find_table_information_separator(dict_rows)

        column_separator_pos = (
            partition_separator_pos if partition_separator_pos > 0 else table_separator_pos
        )
        logger.debug(
            f"relation={relation}, "
            f"partition_separator_pos = {partition_separator_pos}, "
            f"table_separator_pos = {table_separator_pos}, "
            f"column_separator_pos = {column_separator_pos}"
        )

        # fetch columns info
        rows = self.parse_columns_info(raw_rows, 0, table_separator_pos)

        metadata = {
            col["col_name"].split(":")[0].strip(): col["data_type"].strip()
            for col in raw_rows[table_separator_pos + 1:]
            if col["col_name"] and not col["col_name"].startswith("#") and col["data_type"]
        }

        # strip white spaces
        new_metadata = {}
        for k in metadata:
            new_metadata[k.strip()] = metadata[k].strip() if metadata[k] else ""
        metadata = new_metadata

        return [
            DLIColumn(
                table_database=None,
                table_schema=relation.schema,
                table_name=relation.name,
                table_type=relation.type,
                table_owner=str(metadata.get(KEY_TABLE_OWNER)),
                table_stats=None,
                column=column["col_name"],
                column_index=idx,
                dtype=column["data_type"],
            )
            for idx, column in enumerate(rows)
        ]

    @staticmethod
    def find_partition_information_separator(rows: List[dict]) -> int:
        pos = 0
        ps_keys = [
            "# Partition Information",  # non-iceberg
            "# Partition Transform Information",  # iceberg
        ]
        for row in rows:
            if row["col_name"] and row["col_name"].startswith(tuple(ps_keys)):
                break
            pos += 1
        result = 0 if (pos == len(rows)) else pos
        return result

    @staticmethod
    def find_table_information_separator(rows: List[dict]) -> int:
        """Returns the row number of the table information sections"""
        pos = 0
        for row in rows:
            if row["col_name"] and row["col_name"].strip() == KEY_TABLE_INFORMATION:
                break
            pos += 1
        return pos

    @staticmethod
    def parse_columns_info(raw_rows, start, end):
        # remove comments columns
        valid_rows = [
            (row)
            for row in raw_rows[start:end]
            if row["col_name"] and not row["col_name"].startswith("#") and not row["col_name"] == ""
               and not row["col_name"] == "Not partitioned"
        ]

        # avoid overridding duplicate columns by partition columns
        unique_rows = []
        for row in valid_rows:
            curr_unique_keys = list({row["col_name"]: row for row in unique_rows})
            if row["col_name"] not in curr_unique_keys:
                unique_rows.append(row)
        return unique_rows

    def _get_one_catalog(
            self,
            information_schema: InformationSchema,
            schemas: Set[str],
            manifest: Manifest,
    ) -> agate.Table:
        """Get ONE catalog. Used by get_catalog

        manifest is used to run the method in other context's
        threadself.get_columns_in_relation
        """
        if len(schemas) != 1:
            dbt.exceptions.raise_compiler_error(
                f"Expected only one schema in DLI Spark _get_one_catalog, found " f"{schemas}"
            )

        database = information_schema.database
        schema = list(schemas)[0]

        schema_relation = self.Relation.create(
            database=database,
            schema=schema,
            identifier="",
            quote_policy=self.config.quoting,
        ).without_identifier()

        columns: List[Dict[str, Any]] = []
        for relation in self.list_relations(database, schema):
            logger.debug(f"Getting table schema for relation {relation}")
            columns.extend(self._get_columns_for_catalog(relation))

        if len(columns) > 0:
            text_types = agate_helper.build_type_tester(["table_owner", "table_database"])
        else:
            text_types = []

        return agate.Table.from_object(
            columns,
            column_types=text_types,
        )

# may require more build out to make more user friendly to confer with team and community.
